from __future__ import annotations

import datetime
import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._market_configuration import MarketConfiguration, MarketConfigurationWrite
    from ._plant_shop import PlantShop, PlantShopWrite
    from ._shop_result_price_prod import SHOPResultPriceProd, SHOPResultPriceProdWrite


__all__ = [
    "ShopPartialBidCalculationInput",
    "ShopPartialBidCalculationInputWrite",
    "ShopPartialBidCalculationInputApply",
    "ShopPartialBidCalculationInputList",
    "ShopPartialBidCalculationInputWriteList",
    "ShopPartialBidCalculationInputApplyList",
    "ShopPartialBidCalculationInputFields",
    "ShopPartialBidCalculationInputTextFields",
]


ShopPartialBidCalculationInputTextFields = Literal["process_id", "function_name", "function_call_id"]
ShopPartialBidCalculationInputFields = Literal[
    "process_id", "process_step", "function_name", "function_call_id", "step_enabled", "bid_date"
]

_SHOPPARTIALBIDCALCULATIONINPUT_PROPERTIES_BY_FIELD = {
    "process_id": "processId",
    "process_step": "processStep",
    "function_name": "functionName",
    "function_call_id": "functionCallId",
    "step_enabled": "stepEnabled",
    "bid_date": "bidDate",
}


class ShopPartialBidCalculationInput(DomainModel):
    """This represents the reading version of shop partial bid calculation input.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop partial bid calculation input.
        data_record: The data record of the shop partial bid calculation input node.
        process_id: The process associated with the function execution
        process_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        plant: The plant to calculate the partial bid for. Extract price/prod timeseries from Shop Results
        market_configuration: The market configuration to be used to generate the partial bid matrix
        step_enabled: Whether the step is enabled or not
        bid_date: The bid date
        shop_result_price_prod: An array of shop results with price/prod timeserires pairs for all plants included in the respective shop scenario
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "ShopPartialBidCalculationInput"
    )
    process_id: str = Field(alias="processId")
    process_step: int = Field(alias="processStep")
    function_name: str = Field(alias="functionName")
    function_call_id: str = Field(alias="functionCallId")
    plant: Union[PlantShop, str, dm.NodeId, None] = Field(None, repr=False)
    market_configuration: Union[MarketConfiguration, str, dm.NodeId, None] = Field(
        None, repr=False, alias="marketConfiguration"
    )
    step_enabled: Optional[bool] = Field(None, alias="stepEnabled")
    bid_date: Optional[datetime.date] = Field(None, alias="bidDate")
    shop_result_price_prod: Union[list[SHOPResultPriceProd], list[str], None] = Field(
        default=None, repr=False, alias="shopResultPriceProd"
    )

    def as_write(self) -> ShopPartialBidCalculationInputWrite:
        """Convert this read version of shop partial bid calculation input to the writing version."""
        return ShopPartialBidCalculationInputWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            process_id=self.process_id,
            process_step=self.process_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            plant=self.plant.as_write() if isinstance(self.plant, DomainModel) else self.plant,
            market_configuration=(
                self.market_configuration.as_write()
                if isinstance(self.market_configuration, DomainModel)
                else self.market_configuration
            ),
            step_enabled=self.step_enabled,
            bid_date=self.bid_date,
            shop_result_price_prod=[
                (
                    shop_result_price_prod.as_write()
                    if isinstance(shop_result_price_prod, DomainModel)
                    else shop_result_price_prod
                )
                for shop_result_price_prod in self.shop_result_price_prod or []
            ],
        )

    def as_apply(self) -> ShopPartialBidCalculationInputWrite:
        """Convert this read version of shop partial bid calculation input to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopPartialBidCalculationInputWrite(DomainModelWrite):
    """This represents the writing version of shop partial bid calculation input.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop partial bid calculation input.
        data_record: The data record of the shop partial bid calculation input node.
        process_id: The process associated with the function execution
        process_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        plant: The plant to calculate the partial bid for. Extract price/prod timeseries from Shop Results
        market_configuration: The market configuration to be used to generate the partial bid matrix
        step_enabled: Whether the step is enabled or not
        bid_date: The bid date
        shop_result_price_prod: An array of shop results with price/prod timeserires pairs for all plants included in the respective shop scenario
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "ShopPartialBidCalculationInput"
    )
    process_id: str = Field(alias="processId")
    process_step: int = Field(alias="processStep")
    function_name: str = Field(alias="functionName")
    function_call_id: str = Field(alias="functionCallId")
    plant: Union[PlantShopWrite, str, dm.NodeId, None] = Field(None, repr=False)
    market_configuration: Union[MarketConfigurationWrite, str, dm.NodeId, None] = Field(
        None, repr=False, alias="marketConfiguration"
    )
    step_enabled: Optional[bool] = Field(None, alias="stepEnabled")
    bid_date: Optional[datetime.date] = Field(None, alias="bidDate")
    shop_result_price_prod: Union[list[SHOPResultPriceProdWrite], list[str], None] = Field(
        default=None, repr=False, alias="shopResultPriceProd"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            ShopPartialBidCalculationInput, dm.ViewId("sp_powerops_models", "ShopPartialBidCalculationInput", "1")
        )

        properties: dict[str, Any] = {}

        if self.process_id is not None:
            properties["processId"] = self.process_id

        if self.process_step is not None:
            properties["processStep"] = self.process_step

        if self.function_name is not None:
            properties["functionName"] = self.function_name

        if self.function_call_id is not None:
            properties["functionCallId"] = self.function_call_id

        if self.plant is not None:
            properties["plant"] = {
                "space": self.space if isinstance(self.plant, str) else self.plant.space,
                "externalId": self.plant if isinstance(self.plant, str) else self.plant.external_id,
            }

        if self.market_configuration is not None:
            properties["marketConfiguration"] = {
                "space": self.space if isinstance(self.market_configuration, str) else self.market_configuration.space,
                "externalId": (
                    self.market_configuration
                    if isinstance(self.market_configuration, str)
                    else self.market_configuration.external_id
                ),
            }

        if self.step_enabled is not None or write_none:
            properties["stepEnabled"] = self.step_enabled

        if self.bid_date is not None or write_none:
            properties["bidDate"] = self.bid_date.isoformat() if self.bid_date else None

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "SHOPResultPriceProd")
        for shop_result_price_prod in self.shop_result_price_prod or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=shop_result_price_prod,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
            )
            resources.extend(other_resources)

        if isinstance(self.plant, DomainModelWrite):
            other_resources = self.plant._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.market_configuration, DomainModelWrite):
            other_resources = self.market_configuration._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class ShopPartialBidCalculationInputApply(ShopPartialBidCalculationInputWrite):
    def __new__(cls, *args, **kwargs) -> ShopPartialBidCalculationInputApply:
        warnings.warn(
            "ShopPartialBidCalculationInputApply is deprecated and will be removed in v1.0. Use ShopPartialBidCalculationInputWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "ShopPartialBidCalculationInput.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class ShopPartialBidCalculationInputList(DomainModelList[ShopPartialBidCalculationInput]):
    """List of shop partial bid calculation inputs in the read version."""

    _INSTANCE = ShopPartialBidCalculationInput

    def as_write(self) -> ShopPartialBidCalculationInputWriteList:
        """Convert these read versions of shop partial bid calculation input to the writing versions."""
        return ShopPartialBidCalculationInputWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> ShopPartialBidCalculationInputWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopPartialBidCalculationInputWriteList(DomainModelWriteList[ShopPartialBidCalculationInputWrite]):
    """List of shop partial bid calculation inputs in the writing version."""

    _INSTANCE = ShopPartialBidCalculationInputWrite


class ShopPartialBidCalculationInputApplyList(ShopPartialBidCalculationInputWriteList): ...


def _create_shop_partial_bid_calculation_input_filter(
    view_id: dm.ViewId,
    process_id: str | list[str] | None = None,
    process_id_prefix: str | None = None,
    min_process_step: int | None = None,
    max_process_step: int | None = None,
    function_name: str | list[str] | None = None,
    function_name_prefix: str | None = None,
    function_call_id: str | list[str] | None = None,
    function_call_id_prefix: str | None = None,
    plant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    market_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    step_enabled: bool | None = None,
    min_bid_date: datetime.date | None = None,
    max_bid_date: datetime.date | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(process_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("processId"), value=process_id))
    if process_id and isinstance(process_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("processId"), values=process_id))
    if process_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("processId"), value=process_id_prefix))
    if min_process_step is not None or max_process_step is not None:
        filters.append(
            dm.filters.Range(view_id.as_property_ref("processStep"), gte=min_process_step, lte=max_process_step)
        )
    if isinstance(function_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionName"), value=function_name))
    if function_name and isinstance(function_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionName"), values=function_name))
    if function_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionName"), value=function_name_prefix))
    if isinstance(function_call_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionCallId"), value=function_call_id))
    if function_call_id and isinstance(function_call_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionCallId"), values=function_call_id))
    if function_call_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionCallId"), value=function_call_id_prefix))
    if plant and isinstance(plant, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("plant"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": plant}
            )
        )
    if plant and isinstance(plant, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("plant"), value={"space": plant[0], "externalId": plant[1]})
        )
    if plant and isinstance(plant, list) and isinstance(plant[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("plant"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in plant],
            )
        )
    if plant and isinstance(plant, list) and isinstance(plant[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("plant"), values=[{"space": item[0], "externalId": item[1]} for item in plant]
            )
        )
    if market_configuration and isinstance(market_configuration, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("marketConfiguration"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": market_configuration},
            )
        )
    if market_configuration and isinstance(market_configuration, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("marketConfiguration"),
                value={"space": market_configuration[0], "externalId": market_configuration[1]},
            )
        )
    if market_configuration and isinstance(market_configuration, list) and isinstance(market_configuration[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("marketConfiguration"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in market_configuration],
            )
        )
    if market_configuration and isinstance(market_configuration, list) and isinstance(market_configuration[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("marketConfiguration"),
                values=[{"space": item[0], "externalId": item[1]} for item in market_configuration],
            )
        )
    if isinstance(step_enabled, bool):
        filters.append(dm.filters.Equals(view_id.as_property_ref("stepEnabled"), value=step_enabled))
    if min_bid_date is not None or max_bid_date is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("bidDate"),
                gte=min_bid_date.isoformat() if min_bid_date else None,
                lte=max_bid_date.isoformat() if max_bid_date else None,
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from typing import Literal, Optional

import arrow
from cognite.client.data_classes import Event, FileMetadata
from pydantic import BaseModel, ConfigDict, field_validator
from typing_extensions import Self

_SHOP_VERSION_FALLBACK = "15.3.3.2"


class DayaheadFunctionEvent:
    event_type: str = "POWEROPS_FUNCTION_CALL"
    external_id_prefix: str = "POWEROPS_FUNCTION_CALL_"
    function_external_id: str = "function_external_id"
    bid_date: str = "bid:date"
    workflow_event_external_id: str = "workflow_event_external_id"

    @classmethod
    def as_cdf_event(
        cls,
        data_set: int,
        workflow_event_external_id: str,
        bid_date: str,
        event_external_id: str,
        function_name: str,
        additional_metadata: dict,
    ) -> Event:
        return Event(
            external_id=event_external_id,
            type=PartialFunctionEvent.event_type,
            data_set_id=data_set,
            metadata={
                **additional_metadata,
                cls.bid_date: bid_date,
                cls.workflow_event_external_id: workflow_event_external_id,
                cls.function_external_id: function_name,
                "process_type": DayaheadFunctionEvent.event_type,
                "processed": "true",
            },
        )


class PartialFunctionEvent(DayaheadFunctionEvent):
    plant: str = "bid:plant"
    method: str = "bid:bid_matrix_generation_method"
    relationship_label_to_trigger_event: str = "relationship_to.bid_process_event"
    # This is a temporary hack that is introduced due
    # to the potential inconsistency between plant names in
    # SHOP files and the plant asset names in CDF
    plant_name_override: str = "bid:plant_name_override"


class TotalFunctionEvent(DayaheadFunctionEvent):
    portfolio: str = "bid:portfolio"
    bid_process_configuration_name: str = "bid:bid_process_configuration_name"
    relationship_label_to_trigger_event: str = "relationship_to.calculate_total_bid_matrix_event"


class DayaheadTriggerEvent:
    event_type: str = "POWEROPS_BID_PROCESS_FROM_PRERUNS"
    external_id_prefix: str = "POWEROPS_BID_PROCESS_"
    market: str = "bid:market"
    main_scenario: str = "bid:main_scenario"
    market_configuration_nordpool_dayahead: str = "bid:market_config_external_id"
    combine_bid_matrix_tasks: str = "bid:combine_bid_matrix_tasks"
    bid_process_configuration_name: str = "bid:bid_process_configuration_name"
    bid_matrix_generator_config_external_id: str = "bid:bid_matrix_generator_config_external_id"
    price_scenarios: str = "bid:price_scenarios"
    method: str = "bid:bid_process_configuration_name"
    price_area: str = "bid:price_area"
    bid_date: str = "bid:date"
    start_time: str = "shop:starttime"
    end_time: str = "shop:endtime"
    relationship_label_to_shop_run_events: str = "relationship_to.shop_run_event"

    @classmethod
    def as_cdf_event(
        cls,
        event_external_id: str,
        data_set: int,
        start_time: str,
        end_time: str,
        bid_date: str,
        price_scenarios: list[str],
        price_area: str,
        market_configuration_nordpool_dayahead: str,
        bid_configuration_name,
        main_scenario: str = "",
    ) -> Event:
        return Event(
            external_id=event_external_id,
            type=DayaheadTriggerEvent.event_type,
            data_set_id=data_set,
            metadata={
                cls.market: "Dayahead",
                cls.main_scenario: main_scenario,
                cls.price_scenarios: ",".join(price_scenarios),
                cls.bid_date: bid_date,
                cls.start_time: start_time,
                cls.end_time: end_time,
                cls.market_configuration_nordpool_dayahead: market_configuration_nordpool_dayahead,
                cls.bid_process_configuration_name: bid_configuration_name,
                cls.bid_matrix_generator_config_external_id: f"POWEROPS_bid_matrix_generator_config_"
                f"{bid_configuration_name}",
                cls.price_area: f"price_area_{price_area}",
                cls.combine_bid_matrix_tasks: "true",
                "processed": "true",
            },
        )


class PrerunFileMetadata:
    plants: str = "shop:plants"
    price_scenario: str = "shop:price_scenario"
    _plants_delimiter: str = ","


class ShopPreRunFile(BaseModel):
    """
    Represents a single shop run based on a pre-run file.
    """

    pre_run_external_id: str
    plants: list[str]
    price_scenario: str

    @classmethod
    def load_from_metadata(cls, file_metadata: FileMetadata) -> Self:
        plants_as_string = file_metadata.metadata.get(PrerunFileMetadata.plants)
        plants = [plant.lstrip() for plant in plants_as_string.split(PrerunFileMetadata._plants_delimiter)]
        return cls(
            pre_run_external_id=file_metadata.external_id,
            plants=plants,
            price_scenario=file_metadata.metadata.get(PrerunFileMetadata.price_scenario),
        )


class Case(BaseModel):
    """
    Case definition based on a set of prerun files to run for certain plants that are used in the set of prerun files

    """

    case_name: str  # e.g. Glomma
    pre_run_file_external_ids: list[str]
    commands_file: Optional[str] = None
    pre_runs_external_id_prefix: Optional[str] = None


class BidTimeFrame(BaseModel):
    """
    Used to dynamically specify what times to run the Dayahead bid process for in local time

    Args:
        shift_start_in_days: number of days from the current local time to shift the start time of the bid
        timezone: the local timezone to use when creating shop times
        bid_period_in_days: the number of days the bid is valid for. This number is used to produce the bid end time by
                            shifting the start time with this number
    """

    model_config = ConfigDict(frozen=True)

    shift_start_in_days: int
    timezone: Optional[str] = "Europe/Oslo"
    bid_period_in_days: Optional[int] = 14
    _datetime_string_format: str = "YYYY-MM-DD HH:mm:ss"
    _date_string_format: str = "YYYY-MM-DD"
    _start_time_arrow: arrow.Arrow

    @field_validator("bid_period_in_days")
    @classmethod
    def positive_bid_period_in_days(cls, value):
        if value < 0:
            raise ValueError("Bid period must be a positive number")

    @property
    def start_time_arrow(self):
        return (
            arrow.now(self.timezone).shift(days=self.shift_start_in_days).floor("day")
        )  # should floor be hour instead?

    @property
    def start_time_string(self) -> str:
        return self.start_time_arrow.format(self._datetime_string_format)

    @property
    def end_time_string(self) -> str:
        return (
            self.start_time_arrow.shift(days=self.bid_period_in_days).floor("week").format(self._datetime_string_format)
        )

    @property
    def bid_date(self) -> str:
        return self.start_time_arrow.shift(days=1).format(self._date_string_format)


class DayaheadTrigger(BaseModel):
    """
    Contains the blueprint of a Dayahead workflow definition. Used to trigger one or more cases within a price area
    along with SHOP prerun files (e.g. no preprocessing to the SHOP file before being run in CogShop)
    This object is used together with the powerops client (DayAheadTriggerAPI) to trigger a workflow from an instance
    of this class.

    Args:
        price_scenarios: list of price scenarios used in Dayahead run (do not have to exist in CDF, only appended as
                         metadata to the triggering events for reference)
        main_scenario: Main price scenario
        shop_version: Shop Version to use in the run. The Shop version must be installed to Cogshop or exist as a
                      binary file in CDF. For Mesh files specify a SHOP version > 15.4
        price_area: str
        method: Literal["multi_scenario", "price_independent"]
        bid_configuration_name: str
        cases: list of Case objects to run
        bid_time_frame: Optional[BidTimeFrame] = BidTimeFrame(shift_start_in_days=0)
        plant_names_override: Optional[dict] = None
        dayahead_configuration_external_id: Optional[str] = "market_configuration_nordpool_dayahead"
    """

    price_scenarios: list[str]
    main_scenario: str = ""
    shop_version: Optional[str] = _SHOP_VERSION_FALLBACK
    price_area: str
    method: Literal["multi_scenario", "price_independent"]
    bid_configuration_name: str
    cases: list[Case]
    bid_time_frame: Optional[BidTimeFrame] = BidTimeFrame(shift_start_in_days=0)
    plant_names_override: Optional[dict] = None
    dayahead_configuration_external_id: Optional[str] = "market_configuration_nordpool_dayahead"


class DayaheadWorkflowRun(BaseModel):
    """
    Return object when triggering a DayaheadTrigger workflow via the API.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    workflow_trigger_event: Event
    total_bid_event: Event
    partial_bid_events: list[Event]
    shop_run_events: list[Event]



from __future__ import annotations

from cognite.powerops.resync.config import PriceScenario, PriceScenarioID, PriceScenarioV2


def _map_price_scenarios_by_name(
    scenario_ids: list[PriceScenarioID], price_scenarios_by_id: dict[str, PriceScenario], market_name: str
) -> dict[str, PriceScenario]:
    scenario_by_name = {}
    for identifier in scenario_ids:
        ref_scenario = price_scenarios_by_id[identifier.id]
        name = identifier.rename or ref_scenario.name or identifier.id
        scenario_by_name[name] = PriceScenario(name=market_name, **ref_scenario.model_dump(exclude={"name"}))
    return scenario_by_name


def _map_price_scenario_by_name_v2(
    scenario_ids: list[PriceScenarioID], price_scenarios_by_id: dict[str, PriceScenarioV2], market_name: str
) -> dict[str, PriceScenarioV2]:
    scenario_by_name = {}
    for identifier in scenario_ids:
        ref_scenario = price_scenarios_by_id[identifier.id]
        name = identifier.rename or ref_scenario.name or identifier.id
        scenario_by_name[name] = PriceScenario(name=market_name, **ref_scenario.model_dump(exclude={"name"}))
    return scenario_by_name



from __future__ import annotations


from cognite.client import data_modeling as dm

from ._core import DEFAULT_LIMIT_READ, EdgeAPI, _create_edge_filter
from cognite.powerops.client._generated.assets.data_classes._core import DEFAULT_INSTANCE_SPACE


class GeneratorTurbineCurvesAPI(EdgeAPI):
    def list(
        self,
        from_generator: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
        from_generator_space: str = DEFAULT_INSTANCE_SPACE,
        to_turbine_efficiency_curve: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
        to_turbine_efficiency_curve_space: str = DEFAULT_INSTANCE_SPACE,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit=DEFAULT_LIMIT_READ,
    ) -> dm.EdgeList:
        """List turbine curve edges of a generator.

        Args:
            from_generator: ID of the source generator.
            from_generator_space: Location of the generators.
            to_turbine_efficiency_curve: ID of the target turbine efficiency curve.
            to_turbine_efficiency_curve_space: Location of the turbine efficiency curves.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of turbine curve edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            The requested turbine curve edges.

        Examples:

            List 5 turbine curve edges connected to "my_generator":

                >>> from cognite.powerops.client._generated.assets import PowerAssetAPI
                >>> client = PowerAssetAPI()
                >>> generator = client.generator.turbine_curves_edge.list("my_generator", limit=5)

        """
        filter_ = _create_edge_filter(
            dm.DirectRelationReference("power-ops-types", "isSubAssetOf"),
            from_generator,
            from_generator_space,
            to_turbine_efficiency_curve,
            to_turbine_efficiency_curve_space,
            external_id_prefix,
            space,
        )
        return self._list(filter_=filter_, limit=limit)



from __future__ import annotations

import warnings
from pathlib import Path
from typing import Sequence

from cognite.client import ClientConfig, CogniteClient, data_modeling as dm
from cognite.client.data_classes import TimeSeriesList
from cognite.client.credentials import OAuthClientCredentials

from ._api.alert import AlertAPI
from ._api.basic_bid_matrix import BasicBidMatrixAPI
from ._api.bid_calculation_task import BidCalculationTaskAPI
from ._api.bid_configuration import BidConfigurationAPI
from ._api.bid_configuration_shop import BidConfigurationShopAPI
from ._api.bid_configuration_water import BidConfigurationWaterAPI
from ._api.bid_document_afrr import BidDocumentAFRRAPI
from ._api.bid_document_day_ahead import BidDocumentDayAheadAPI
from ._api.bid_matrix import BidMatrixAPI
from ._api.bid_matrix_raw import BidMatrixRawAPI
from ._api.bid_method import BidMethodAPI
from ._api.bid_method_afrr import BidMethodAFRRAPI
from ._api.bid_method_custom import BidMethodCustomAPI
from ._api.bid_method_day_ahead import BidMethodDayAheadAPI
from ._api.bid_method_shop_multi_scenario import BidMethodSHOPMultiScenarioAPI
from ._api.bid_method_water_value import BidMethodWaterValueAPI
from ._api.bid_row import BidRowAPI
from ._api.case import CaseAPI
from ._api.commands import CommandsAPI
from ._api.custom_bid_matrix import CustomBidMatrixAPI
from ._api.generator import GeneratorAPI
from ._api.generator_efficiency_curve import GeneratorEfficiencyCurveAPI
from ._api.mapping import MappingAPI
from ._api.market_configuration import MarketConfigurationAPI
from ._api.model_template import ModelTemplateAPI
from ._api.multi_scenario_matrix import MultiScenarioMatrixAPI
from ._api.multi_scenario_matrix_raw import MultiScenarioMatrixRawAPI
from ._api.partial_post_processing_input import PartialPostProcessingInputAPI
from ._api.partial_post_processing_output import PartialPostProcessingOutputAPI
from ._api.plant import PlantAPI
from ._api.plant_shop import PlantShopAPI
from ._api.preprocessor_input import PreprocessorInputAPI
from ._api.preprocessor_output import PreprocessorOutputAPI
from ._api.price_area import PriceAreaAPI
from ._api.price_area_afrr import PriceAreaAFRRAPI
from ._api.price_area_asset import PriceAreaAssetAPI
from ._api.price_prod_case import PriceProdCaseAPI
from ._api.reservoir import ReservoirAPI
from ._api.shop_result import SHOPResultAPI
from ._api.shop_result_price_prod import SHOPResultPriceProdAPI
from ._api.shop_time_series import SHOPTimeSeriesAPI
from ._api.shop_trigger_input import SHOPTriggerInputAPI
from ._api.shop_trigger_output import SHOPTriggerOutputAPI
from ._api.scenario import ScenarioAPI
from ._api.shop_partial_bid_calculation_input import ShopPartialBidCalculationInputAPI
from ._api.shop_partial_bid_calculation_output import ShopPartialBidCalculationOutputAPI
from ._api.task_dispatcher_shop_input import TaskDispatcherShopInputAPI
from ._api.task_dispatcher_shop_output import TaskDispatcherShopOutputAPI
from ._api.task_dispatcher_water_input import TaskDispatcherWaterInputAPI
from ._api.task_dispatcher_water_output import TaskDispatcherWaterOutputAPI
from ._api.total_bid_matrix_calculation_input import TotalBidMatrixCalculationInputAPI
from ._api.total_bid_matrix_calculation_output import TotalBidMatrixCalculationOutputAPI
from ._api.turbine_efficiency_curve import TurbineEfficiencyCurveAPI
from ._api.water_partial_bid_calculation_input import WaterPartialBidCalculationInputAPI
from ._api.water_partial_bid_calculation_output import WaterPartialBidCalculationOutputAPI
from ._api.watercourse import WatercourseAPI
from ._api.watercourse_shop import WatercourseShopAPI
from ._api._core import SequenceNotStr
from .data_classes._core import DEFAULT_INSTANCE_SPACE
from . import data_classes


class SHOPBasedDayAheadBidProcesAPIs:
    """
    SHOPBasedDayAheadBidProcesAPIs

    Data Model:
        space: sp_powerops_models
        externalId: compute_SHOPBasedDayAhead
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.Alert: dm.ViewId("sp_powerops_models", "Alert", "1"),
            data_classes.BidConfigurationShop: dm.ViewId("sp_powerops_models", "BidConfigurationShop", "1"),
            data_classes.BidMatrixRaw: dm.ViewId("sp_powerops_models", "BidMatrixRaw", "1"),
            data_classes.BidMethodSHOPMultiScenario: dm.ViewId("sp_powerops_models", "BidMethodSHOPMultiScenario", "1"),
            data_classes.Case: dm.ViewId("sp_powerops_models", "Case", "1"),
            data_classes.Commands: dm.ViewId("sp_powerops_models", "Commands", "1"),
            data_classes.Mapping: dm.ViewId("sp_powerops_models", "Mapping", "1"),
            data_classes.MarketConfiguration: dm.ViewId("sp_powerops_models", "MarketConfiguration", "1"),
            data_classes.ModelTemplate: dm.ViewId("sp_powerops_models", "ModelTemplate", "1"),
            data_classes.MultiScenarioMatrixRaw: dm.ViewId("sp_powerops_models", "MultiScenarioMatrixRaw", "1"),
            data_classes.PlantShop: dm.ViewId("sp_powerops_models", "PlantShop", "1"),
            data_classes.PreprocessorInput: dm.ViewId("sp_powerops_models", "PreprocessorInput", "1"),
            data_classes.PreprocessorOutput: dm.ViewId("sp_powerops_models", "PreprocessorOutput", "1"),
            data_classes.PriceArea: dm.ViewId("sp_powerops_models", "PriceArea", "1"),
            data_classes.PriceProdCase: dm.ViewId("sp_powerops_models", "PriceProdCase", "1"),
            data_classes.SHOPResult: dm.ViewId("sp_powerops_models", "SHOPResult", "1"),
            data_classes.SHOPResultPriceProd: dm.ViewId("sp_powerops_models", "SHOPResultPriceProd", "1"),
            data_classes.SHOPTimeSeries: dm.ViewId("sp_powerops_models", "SHOPTimeSeries", "1"),
            data_classes.SHOPTriggerInput: dm.ViewId("sp_powerops_models", "SHOPTriggerInput", "1"),
            data_classes.SHOPTriggerOutput: dm.ViewId("sp_powerops_models", "SHOPTriggerOutput", "1"),
            data_classes.Scenario: dm.ViewId("sp_powerops_models", "Scenario", "1"),
            data_classes.ShopPartialBidCalculationInput: dm.ViewId(
                "sp_powerops_models", "ShopPartialBidCalculationInput", "1"
            ),
            data_classes.ShopPartialBidCalculationOutput: dm.ViewId(
                "sp_powerops_models", "ShopPartialBidCalculationOutput", "1"
            ),
            data_classes.TaskDispatcherShopInput: dm.ViewId("sp_powerops_models", "TaskDispatcherShopInput", "1"),
            data_classes.TaskDispatcherShopOutput: dm.ViewId("sp_powerops_models", "TaskDispatcherShopOutput", "1"),
            data_classes.WatercourseShop: dm.ViewId("sp_powerops_models", "WatercourseShop", "1"),
        }
        self._view_by_read_class = view_by_read_class

        self.alert = AlertAPI(client, view_by_read_class)
        self.bid_configuration_shop = BidConfigurationShopAPI(client, view_by_read_class)
        self.bid_matrix_raw = BidMatrixRawAPI(client, view_by_read_class)
        self.bid_method_shop_multi_scenario = BidMethodSHOPMultiScenarioAPI(client, view_by_read_class)
        self.case = CaseAPI(client, view_by_read_class)
        self.commands = CommandsAPI(client, view_by_read_class)
        self.mapping = MappingAPI(client, view_by_read_class)
        self.market_configuration = MarketConfigurationAPI(client, view_by_read_class)
        self.model_template = ModelTemplateAPI(client, view_by_read_class)
        self.multi_scenario_matrix_raw = MultiScenarioMatrixRawAPI(client, view_by_read_class)
        self.plant_shop = PlantShopAPI(client, view_by_read_class)
        self.preprocessor_input = PreprocessorInputAPI(client, view_by_read_class)
        self.preprocessor_output = PreprocessorOutputAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.price_prod_case = PriceProdCaseAPI(client, view_by_read_class)
        self.shop_result = SHOPResultAPI(client, view_by_read_class)
        self.shop_result_price_prod = SHOPResultPriceProdAPI(client, view_by_read_class)
        self.shop_time_series = SHOPTimeSeriesAPI(client, view_by_read_class)
        self.shop_trigger_input = SHOPTriggerInputAPI(client, view_by_read_class)
        self.shop_trigger_output = SHOPTriggerOutputAPI(client, view_by_read_class)
        self.scenario = ScenarioAPI(client, view_by_read_class)
        self.shop_partial_bid_calculation_input = ShopPartialBidCalculationInputAPI(client, view_by_read_class)
        self.shop_partial_bid_calculation_output = ShopPartialBidCalculationOutputAPI(client, view_by_read_class)
        self.task_dispatcher_shop_input = TaskDispatcherShopInputAPI(client, view_by_read_class)
        self.task_dispatcher_shop_output = TaskDispatcherShopOutputAPI(client, view_by_read_class)
        self.watercourse_shop = WatercourseShopAPI(client, view_by_read_class)


class TotalBidCalculationAPIs:
    """
    TotalBidCalculationAPIs

    Data Model:
        space: sp_powerops_models
        externalId: compute_TotalBidCalculation
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.Alert: dm.ViewId("sp_powerops_models", "Alert", "1"),
            data_classes.BidDocumentDayAhead: dm.ViewId("sp_powerops_models", "BidDocumentDayAhead", "1"),
            data_classes.BidMatrix: dm.ViewId("sp_powerops_models", "BidMatrix", "1"),
            data_classes.BidMatrixRaw: dm.ViewId("sp_powerops_models", "BidMatrixRaw", "1"),
            data_classes.BidMethodDayAhead: dm.ViewId("sp_powerops_models", "BidMethodDayAhead", "1"),
            data_classes.BidMethodSHOPMultiScenario: dm.ViewId("sp_powerops_models", "BidMethodSHOPMultiScenario", "1"),
            data_classes.BidMethodWaterValue: dm.ViewId("sp_powerops_models", "BidMethodWaterValue", "1"),
            data_classes.Case: dm.ViewId("sp_powerops_models", "Case", "1"),
            data_classes.Commands: dm.ViewId("sp_powerops_models", "Commands", "1"),
            data_classes.Mapping: dm.ViewId("sp_powerops_models", "Mapping", "1"),
            data_classes.MarketConfiguration: dm.ViewId("sp_powerops_models", "MarketConfiguration", "1"),
            data_classes.ModelTemplate: dm.ViewId("sp_powerops_models", "ModelTemplate", "1"),
            data_classes.MultiScenarioMatrix: dm.ViewId("sp_powerops_models", "MultiScenarioMatrix", "1"),
            data_classes.MultiScenarioMatrixRaw: dm.ViewId("sp_powerops_models", "MultiScenarioMatrixRaw", "1"),
            data_classes.PartialPostProcessingInput: dm.ViewId("sp_powerops_models", "PartialPostProcessingInput", "1"),
            data_classes.PartialPostProcessingOutput: dm.ViewId(
                "sp_powerops_models", "PartialPostProcessingOutput", "1"
            ),
            data_classes.PriceArea: dm.ViewId("sp_powerops_models", "PriceArea", "1"),
            data_classes.PriceProdCase: dm.ViewId("sp_powerops_models", "PriceProdCase", "1"),
            data_classes.SHOPResult: dm.ViewId("sp_powerops_models", "SHOPResult", "1"),
            data_classes.SHOPResultPriceProd: dm.ViewId("sp_powerops_models", "SHOPResultPriceProd", "1"),
            data_classes.SHOPTimeSeries: dm.ViewId("sp_powerops_models", "SHOPTimeSeries", "1"),
            data_classes.Scenario: dm.ViewId("sp_powerops_models", "Scenario", "1"),
            data_classes.TotalBidMatrixCalculationInput: dm.ViewId(
                "sp_powerops_models", "TotalBidMatrixCalculationInput", "1"
            ),
            data_classes.TotalBidMatrixCalculationOutput: dm.ViewId(
                "sp_powerops_models", "TotalBidMatrixCalculationOutput", "1"
            ),
            data_classes.WatercourseShop: dm.ViewId("sp_powerops_models", "WatercourseShop", "1"),
        }
        self._view_by_read_class = view_by_read_class

        self.alert = AlertAPI(client, view_by_read_class)
        self.bid_document_day_ahead = BidDocumentDayAheadAPI(client, view_by_read_class)
        self.bid_matrix = BidMatrixAPI(client, view_by_read_class)
        self.bid_matrix_raw = BidMatrixRawAPI(client, view_by_read_class)
        self.bid_method_day_ahead = BidMethodDayAheadAPI(client, view_by_read_class)
        self.bid_method_shop_multi_scenario = BidMethodSHOPMultiScenarioAPI(client, view_by_read_class)
        self.bid_method_water_value = BidMethodWaterValueAPI(client, view_by_read_class)
        self.case = CaseAPI(client, view_by_read_class)
        self.commands = CommandsAPI(client, view_by_read_class)
        self.mapping = MappingAPI(client, view_by_read_class)
        self.market_configuration = MarketConfigurationAPI(client, view_by_read_class)
        self.model_template = ModelTemplateAPI(client, view_by_read_class)
        self.multi_scenario_matrix = MultiScenarioMatrixAPI(client, view_by_read_class)
        self.multi_scenario_matrix_raw = MultiScenarioMatrixRawAPI(client, view_by_read_class)
        self.partial_post_processing_input = PartialPostProcessingInputAPI(client, view_by_read_class)
        self.partial_post_processing_output = PartialPostProcessingOutputAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.price_prod_case = PriceProdCaseAPI(client, view_by_read_class)
        self.shop_result = SHOPResultAPI(client, view_by_read_class)
        self.shop_result_price_prod = SHOPResultPriceProdAPI(client, view_by_read_class)
        self.shop_time_series = SHOPTimeSeriesAPI(client, view_by_read_class)
        self.scenario = ScenarioAPI(client, view_by_read_class)
        self.total_bid_matrix_calculation_input = TotalBidMatrixCalculationInputAPI(client, view_by_read_class)
        self.total_bid_matrix_calculation_output = TotalBidMatrixCalculationOutputAPI(client, view_by_read_class)
        self.watercourse_shop = WatercourseShopAPI(client, view_by_read_class)


class WaterValueBasedDayAheadBidProcesAPIs:
    """
    WaterValueBasedDayAheadBidProcesAPIs

    Data Model:
        space: sp_powerops_models
        externalId: compute_WaterValueBasedDayAheadBid
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.Alert: dm.ViewId("sp_powerops_models", "Alert", "1"),
            data_classes.BidCalculationTask: dm.ViewId("sp_powerops_models", "BidCalculationTask", "1"),
            data_classes.BidConfigurationWater: dm.ViewId("sp_powerops_models", "BidConfigurationWater", "1"),
            data_classes.BidMatrixRaw: dm.ViewId("sp_powerops_models", "BidMatrixRaw", "1"),
            data_classes.BidMethodWaterValue: dm.ViewId("sp_powerops_models", "BidMethodWaterValue", "1"),
            data_classes.Generator: dm.ViewId("sp_powerops_models", "Generator", "1"),
            data_classes.GeneratorEfficiencyCurve: dm.ViewId("sp_powerops_models", "GeneratorEfficiencyCurve", "1"),
            data_classes.MarketConfiguration: dm.ViewId("sp_powerops_models", "MarketConfiguration", "1"),
            data_classes.Plant: dm.ViewId("sp_powerops_models", "Plant", "1"),
            data_classes.PriceArea: dm.ViewId("sp_powerops_models", "PriceArea", "1"),
            data_classes.Reservoir: dm.ViewId("sp_powerops_models", "Reservoir", "1"),
            data_classes.TaskDispatcherWaterInput: dm.ViewId("sp_powerops_models", "TaskDispatcherWaterInput", "1"),
            data_classes.TaskDispatcherWaterOutput: dm.ViewId("sp_powerops_models", "TaskDispatcherWaterOutput", "1"),
            data_classes.TurbineEfficiencyCurve: dm.ViewId("sp_powerops_models", "TurbineEfficiencyCurve", "1"),
            data_classes.WaterPartialBidCalculationInput: dm.ViewId(
                "sp_powerops_models", "WaterPartialBidCalculationInput", "1"
            ),
            data_classes.WaterPartialBidCalculationOutput: dm.ViewId(
                "sp_powerops_models", "WaterPartialBidCalculationOutput", "1"
            ),
            data_classes.Watercourse: dm.ViewId("sp_powerops_models", "Watercourse", "1"),
        }
        self._view_by_read_class = view_by_read_class

        self.alert = AlertAPI(client, view_by_read_class)
        self.bid_calculation_task = BidCalculationTaskAPI(client, view_by_read_class)
        self.bid_configuration_water = BidConfigurationWaterAPI(client, view_by_read_class)
        self.bid_matrix_raw = BidMatrixRawAPI(client, view_by_read_class)
        self.bid_method_water_value = BidMethodWaterValueAPI(client, view_by_read_class)
        self.generator = GeneratorAPI(client, view_by_read_class)
        self.generator_efficiency_curve = GeneratorEfficiencyCurveAPI(client, view_by_read_class)
        self.market_configuration = MarketConfigurationAPI(client, view_by_read_class)
        self.plant = PlantAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.reservoir = ReservoirAPI(client, view_by_read_class)
        self.task_dispatcher_water_input = TaskDispatcherWaterInputAPI(client, view_by_read_class)
        self.task_dispatcher_water_output = TaskDispatcherWaterOutputAPI(client, view_by_read_class)
        self.turbine_efficiency_curve = TurbineEfficiencyCurveAPI(client, view_by_read_class)
        self.water_partial_bid_calculation_input = WaterPartialBidCalculationInputAPI(client, view_by_read_class)
        self.water_partial_bid_calculation_output = WaterPartialBidCalculationOutputAPI(client, view_by_read_class)
        self.watercourse = WatercourseAPI(client, view_by_read_class)


class DayAheadConfigurationAPIs:
    """
    DayAheadConfigurationAPIs

    Data Model:
        space: sp_powerops_models
        externalId: config_DayAheadConfiguration
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.BidConfiguration: dm.ViewId("sp_powerops_models", "BidConfiguration", "1"),
            data_classes.BidConfigurationShop: dm.ViewId("sp_powerops_models", "BidConfigurationShop", "1"),
            data_classes.BidConfigurationWater: dm.ViewId("sp_powerops_models", "BidConfigurationWater", "1"),
            data_classes.BidMethod: dm.ViewId("sp_powerops_models", "BidMethod", "1"),
            data_classes.BidMethodDayAhead: dm.ViewId("sp_powerops_models", "BidMethodDayAhead", "1"),
            data_classes.BidMethodSHOPMultiScenario: dm.ViewId("sp_powerops_models", "BidMethodSHOPMultiScenario", "1"),
            data_classes.BidMethodWaterValue: dm.ViewId("sp_powerops_models", "BidMethodWaterValue", "1"),
            data_classes.Commands: dm.ViewId("sp_powerops_models", "Commands", "1"),
            data_classes.Generator: dm.ViewId("sp_powerops_models", "Generator", "1"),
            data_classes.GeneratorEfficiencyCurve: dm.ViewId("sp_powerops_models", "GeneratorEfficiencyCurve", "1"),
            data_classes.Mapping: dm.ViewId("sp_powerops_models", "Mapping", "1"),
            data_classes.MarketConfiguration: dm.ViewId("sp_powerops_models", "MarketConfiguration", "1"),
            data_classes.ModelTemplate: dm.ViewId("sp_powerops_models", "ModelTemplate", "1"),
            data_classes.Plant: dm.ViewId("sp_powerops_models", "Plant", "1"),
            data_classes.PlantShop: dm.ViewId("sp_powerops_models", "PlantShop", "1"),
            data_classes.PriceArea: dm.ViewId("sp_powerops_models", "PriceArea", "1"),
            data_classes.Reservoir: dm.ViewId("sp_powerops_models", "Reservoir", "1"),
            data_classes.Scenario: dm.ViewId("sp_powerops_models", "Scenario", "1"),
            data_classes.TurbineEfficiencyCurve: dm.ViewId("sp_powerops_models", "TurbineEfficiencyCurve", "1"),
            data_classes.Watercourse: dm.ViewId("sp_powerops_models", "Watercourse", "1"),
            data_classes.WatercourseShop: dm.ViewId("sp_powerops_models", "WatercourseShop", "1"),
        }
        self._view_by_read_class = view_by_read_class

        self.bid_configuration = BidConfigurationAPI(client, view_by_read_class)
        self.bid_configuration_shop = BidConfigurationShopAPI(client, view_by_read_class)
        self.bid_configuration_water = BidConfigurationWaterAPI(client, view_by_read_class)
        self.bid_method = BidMethodAPI(client, view_by_read_class)
        self.bid_method_day_ahead = BidMethodDayAheadAPI(client, view_by_read_class)
        self.bid_method_shop_multi_scenario = BidMethodSHOPMultiScenarioAPI(client, view_by_read_class)
        self.bid_method_water_value = BidMethodWaterValueAPI(client, view_by_read_class)
        self.commands = CommandsAPI(client, view_by_read_class)
        self.generator = GeneratorAPI(client, view_by_read_class)
        self.generator_efficiency_curve = GeneratorEfficiencyCurveAPI(client, view_by_read_class)
        self.mapping = MappingAPI(client, view_by_read_class)
        self.market_configuration = MarketConfigurationAPI(client, view_by_read_class)
        self.model_template = ModelTemplateAPI(client, view_by_read_class)
        self.plant = PlantAPI(client, view_by_read_class)
        self.plant_shop = PlantShopAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.reservoir = ReservoirAPI(client, view_by_read_class)
        self.scenario = ScenarioAPI(client, view_by_read_class)
        self.turbine_efficiency_curve = TurbineEfficiencyCurveAPI(client, view_by_read_class)
        self.watercourse = WatercourseAPI(client, view_by_read_class)
        self.watercourse_shop = WatercourseShopAPI(client, view_by_read_class)


class AFRRBidAPIs:
    """
    AFRRBidAPIs

    Data Model:
        space: sp_powerops_models
        externalId: frontend_AFRRBid
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.Alert: dm.ViewId("sp_powerops_models", "Alert", "1"),
            data_classes.BidDocumentAFRR: dm.ViewId("sp_powerops_models", "BidDocumentAFRR", "1"),
            data_classes.BidMethodAFRR: dm.ViewId("sp_powerops_models", "BidMethodAFRR", "1"),
            data_classes.BidRow: dm.ViewId("sp_powerops_models", "BidRow", "1"),
            data_classes.PriceAreaAFRR: dm.ViewId("sp_powerops_models", "PriceAreaAFRR", "1"),
        }
        self._view_by_read_class = view_by_read_class

        self.alert = AlertAPI(client, view_by_read_class)
        self.bid_document_afrr = BidDocumentAFRRAPI(client, view_by_read_class)
        self.bid_method_afrr = BidMethodAFRRAPI(client, view_by_read_class)
        self.bid_row = BidRowAPI(client, view_by_read_class)
        self.price_area_afrr = PriceAreaAFRRAPI(client, view_by_read_class)


class PowerAssetAPIs:
    """
    PowerAssetAPIs

    Data Model:
        space: sp_powerops_models
        externalId: frontend_Asset
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.BidMethodDayAhead: dm.ViewId("sp_powerops_models", "BidMethodDayAhead", "1"),
            data_classes.Generator: dm.ViewId("sp_powerops_models", "Generator", "1"),
            data_classes.GeneratorEfficiencyCurve: dm.ViewId("sp_powerops_models", "GeneratorEfficiencyCurve", "1"),
            data_classes.Plant: dm.ViewId("sp_powerops_models", "Plant", "1"),
            data_classes.PriceAreaAsset: dm.ViewId("sp_powerops_models", "PriceAreaAsset", "1"),
            data_classes.Reservoir: dm.ViewId("sp_powerops_models", "Reservoir", "1"),
            data_classes.TurbineEfficiencyCurve: dm.ViewId("sp_powerops_models", "TurbineEfficiencyCurve", "1"),
            data_classes.Watercourse: dm.ViewId("sp_powerops_models", "Watercourse", "1"),
        }
        self._view_by_read_class = view_by_read_class

        self.bid_method_day_ahead = BidMethodDayAheadAPI(client, view_by_read_class)
        self.generator = GeneratorAPI(client, view_by_read_class)
        self.generator_efficiency_curve = GeneratorEfficiencyCurveAPI(client, view_by_read_class)
        self.plant = PlantAPI(client, view_by_read_class)
        self.price_area_asset = PriceAreaAssetAPI(client, view_by_read_class)
        self.reservoir = ReservoirAPI(client, view_by_read_class)
        self.turbine_efficiency_curve = TurbineEfficiencyCurveAPI(client, view_by_read_class)
        self.watercourse = WatercourseAPI(client, view_by_read_class)


class DayAheadBidAPIs:
    """
    DayAheadBidAPIs

    Data Model:
        space: sp_powerops_models
        externalId: frontend_DayAheadBid
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.Alert: dm.ViewId("sp_powerops_models", "Alert", "1"),
            data_classes.BasicBidMatrix: dm.ViewId("sp_powerops_models", "BasicBidMatrix", "1"),
            data_classes.BidDocumentDayAhead: dm.ViewId("sp_powerops_models", "BidDocumentDayAhead", "1"),
            data_classes.BidMatrix: dm.ViewId("sp_powerops_models", "BidMatrix", "1"),
            data_classes.BidMethodCustom: dm.ViewId("sp_powerops_models", "BidMethodCustom", "1"),
            data_classes.BidMethodDayAhead: dm.ViewId("sp_powerops_models", "BidMethodDayAhead", "1"),
            data_classes.BidMethodSHOPMultiScenario: dm.ViewId("sp_powerops_models", "BidMethodSHOPMultiScenario", "1"),
            data_classes.BidMethodWaterValue: dm.ViewId("sp_powerops_models", "BidMethodWaterValue", "1"),
            data_classes.Case: dm.ViewId("sp_powerops_models", "Case", "1"),
            data_classes.Commands: dm.ViewId("sp_powerops_models", "Commands", "1"),
            data_classes.CustomBidMatrix: dm.ViewId("sp_powerops_models", "CustomBidMatrix", "1"),
            data_classes.Mapping: dm.ViewId("sp_powerops_models", "Mapping", "1"),
            data_classes.ModelTemplate: dm.ViewId("sp_powerops_models", "ModelTemplate", "1"),
            data_classes.MultiScenarioMatrix: dm.ViewId("sp_powerops_models", "MultiScenarioMatrix", "1"),
            data_classes.PriceArea: dm.ViewId("sp_powerops_models", "PriceArea", "1"),
            data_classes.PriceProdCase: dm.ViewId("sp_powerops_models", "PriceProdCase", "1"),
            data_classes.Scenario: dm.ViewId("sp_powerops_models", "Scenario", "1"),
            data_classes.WatercourseShop: dm.ViewId("sp_powerops_models", "WatercourseShop", "1"),
        }
        self._view_by_read_class = view_by_read_class

        self.alert = AlertAPI(client, view_by_read_class)
        self.basic_bid_matrix = BasicBidMatrixAPI(client, view_by_read_class)
        self.bid_document_day_ahead = BidDocumentDayAheadAPI(client, view_by_read_class)
        self.bid_matrix = BidMatrixAPI(client, view_by_read_class)
        self.bid_method_custom = BidMethodCustomAPI(client, view_by_read_class)
        self.bid_method_day_ahead = BidMethodDayAheadAPI(client, view_by_read_class)
        self.bid_method_shop_multi_scenario = BidMethodSHOPMultiScenarioAPI(client, view_by_read_class)
        self.bid_method_water_value = BidMethodWaterValueAPI(client, view_by_read_class)
        self.case = CaseAPI(client, view_by_read_class)
        self.commands = CommandsAPI(client, view_by_read_class)
        self.custom_bid_matrix = CustomBidMatrixAPI(client, view_by_read_class)
        self.mapping = MappingAPI(client, view_by_read_class)
        self.model_template = ModelTemplateAPI(client, view_by_read_class)
        self.multi_scenario_matrix = MultiScenarioMatrixAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.price_prod_case = PriceProdCaseAPI(client, view_by_read_class)
        self.scenario = ScenarioAPI(client, view_by_read_class)
        self.watercourse_shop = WatercourseShopAPI(client, view_by_read_class)


class PowerOpsModelsV1Client:
    """
    PowerOpsModelsV1Client

    Generated with:
        pygen = 0.99.11
        cognite-sdk = 7.26.0
        pydantic = 2.6.3

    """

    def __init__(self, config_or_client: CogniteClient | ClientConfig):
        if isinstance(config_or_client, CogniteClient):
            client = config_or_client
        elif isinstance(config_or_client, ClientConfig):
            client = CogniteClient(config_or_client)
        else:
            raise ValueError(f"Expected CogniteClient or ClientConfig, got {type(config_or_client)}")
        # The client name is used for aggregated logging of Pygen Usage
        client.config.client_name = "CognitePygen:0.99.11"

        self.shop_based_day_ahead_bid_process = SHOPBasedDayAheadBidProcesAPIs(client)
        self.total_bid_calculation = TotalBidCalculationAPIs(client)
        self.water_value_based_day_ahead_bid_process = WaterValueBasedDayAheadBidProcesAPIs(client)
        self.day_ahead_configuration = DayAheadConfigurationAPIs(client)
        self.afrr_bid = AFRRBidAPIs(client)
        self.power_asset = PowerAssetAPIs(client)
        self.day_ahead_bid = DayAheadBidAPIs(client)

        self._client = client
        self._view_by_read_class = {
            k: v
            for api in [
                self.shop_based_day_ahead_bid_process,
                self.total_bid_calculation,
                self.water_value_based_day_ahead_bid_process,
                self.day_ahead_configuration,
                self.afrr_bid,
                self.power_asset,
                self.day_ahead_bid,
            ]
            for k, v in api._view_by_read_class.items()
        }

    def upsert(
        self,
        items: data_classes.DomainModelWrite | Sequence[data_classes.DomainModelWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> data_classes.ResourcesWriteResult:
        """Add or update (upsert) items.

        Args:
            items: One or more instances of the pygen generated data classes.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method will, by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        """
        if isinstance(items, data_classes.DomainModelWrite):
            instances = items.to_instances_write(self._view_by_read_class, write_none)
        else:
            instances = data_classes.ResourcesWrite()
            cache: set[tuple[str, str]] = set()
            for item in items:
                instances.extend(item._to_instances_write(cache, self._view_by_read_class, write_none))
        result = self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )
        time_series = []
        if instances.time_series:
            time_series = self._client.time_series.upsert(instances.time_series, mode="patch")

        return data_classes.ResourcesWriteResult(result.nodes, result.edges, TimeSeriesList(time_series))

    def apply(
        self,
        items: data_classes.DomainModelWrite | Sequence[data_classes.DomainModelWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> data_classes.ResourcesWriteResult:
        """Add or update (upsert) items.

        Args:
            items: One or more instances of the pygen generated data classes.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method will, by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the instead."
            "The motivation is that .upsert is a more descriptive name for the operation.",
            UserWarning,
            stacklevel=2,
        )
        return self.upsert(items, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more items.

        Args:
            external_id: External id of the item(s) to delete.
            space: The space where all the item(s) are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete item by id:

                >>> from omni import OmniClient
                >>> client = OmniClient()
                >>> client.delete("my_node_external_id")
        """
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @classmethod
    def azure_project(
        cls, tenant_id: str, client_id: str, client_secret: str, cdf_cluster: str, project: str
    ) -> PowerOpsModelsV1Client:
        credentials = OAuthClientCredentials.default_for_azure_ad(tenant_id, client_id, client_secret, cdf_cluster)
        config = ClientConfig.default(project, cdf_cluster, credentials)

        return cls(config)

    @classmethod
    def from_toml(cls, file_path: Path | str, section: str | None = "cognite") -> PowerOpsModelsV1Client:
        import toml

        toml_content = toml.load(file_path)
        if section is not None:
            try:
                toml_content = toml_content[section]
            except KeyError as e:
                raise ValueError(f"Could not find section '{section}' in {file_path}") from e

        return cls.azure_project(**toml_content)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)
from ._price_area import PriceArea, PriceAreaWrite

if TYPE_CHECKING:
    from ._plant import Plant, PlantWrite
    from ._watercourse import Watercourse, WatercourseWrite


__all__ = [
    "PriceAreaAsset",
    "PriceAreaAssetWrite",
    "PriceAreaAssetApply",
    "PriceAreaAssetList",
    "PriceAreaAssetWriteList",
    "PriceAreaAssetApplyList",
    "PriceAreaAssetFields",
    "PriceAreaAssetTextFields",
]


PriceAreaAssetTextFields = Literal["name", "timezone"]
PriceAreaAssetFields = Literal["name", "timezone"]

_PRICEAREAASSET_PROPERTIES_BY_FIELD = {
    "name": "name",
    "timezone": "timezone",
}


class PriceAreaAsset(PriceArea):
    """This represents the reading version of price area asset.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area asset.
        data_record: The data record of the price area asset node.
        name: The name of the price area
        timezone: The timezone of the price area
        plants: An array of associated plants.
        watercourses: An array of associated watercourses.
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("sp_powerops_types", "PriceArea")
    plants: Union[list[Plant], list[str], None] = Field(default=None, repr=False)
    watercourses: Union[list[Watercourse], list[str], None] = Field(default=None, repr=False)

    def as_write(self) -> PriceAreaAssetWrite:
        """Convert this read version of price area asset to the writing version."""
        return PriceAreaAssetWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            timezone=self.timezone,
            plants=[plant.as_write() if isinstance(plant, DomainModel) else plant for plant in self.plants or []],
            watercourses=[
                watercourse.as_write() if isinstance(watercourse, DomainModel) else watercourse
                for watercourse in self.watercourses or []
            ],
        )

    def as_apply(self) -> PriceAreaAssetWrite:
        """Convert this read version of price area asset to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceAreaAssetWrite(PriceAreaWrite):
    """This represents the writing version of price area asset.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area asset.
        data_record: The data record of the price area asset node.
        name: The name of the price area
        timezone: The timezone of the price area
        plants: An array of associated plants.
        watercourses: An array of associated watercourses.
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("sp_powerops_types", "PriceArea")
    plants: Union[list[PlantWrite], list[str], None] = Field(default=None, repr=False)
    watercourses: Union[list[WatercourseWrite], list[str], None] = Field(default=None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            PriceAreaAsset, dm.ViewId("sp_powerops_models", "PriceAreaAsset", "1")
        )

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.timezone is not None:
            properties["timezone"] = self.timezone

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "isPlantOf")
        for plant in self.plants or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache, start_node=self, end_node=plant, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("sp_powerops_types", "isWatercourseOf")
        for watercourse in self.watercourses or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache, start_node=self, end_node=watercourse, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        return resources


class PriceAreaAssetApply(PriceAreaAssetWrite):
    def __new__(cls, *args, **kwargs) -> PriceAreaAssetApply:
        warnings.warn(
            "PriceAreaAssetApply is deprecated and will be removed in v1.0. Use PriceAreaAssetWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "PriceAreaAsset.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class PriceAreaAssetList(DomainModelList[PriceAreaAsset]):
    """List of price area assets in the read version."""

    _INSTANCE = PriceAreaAsset

    def as_write(self) -> PriceAreaAssetWriteList:
        """Convert these read versions of price area asset to the writing versions."""
        return PriceAreaAssetWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> PriceAreaAssetWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceAreaAssetWriteList(DomainModelWriteList[PriceAreaAssetWrite]):
    """List of price area assets in the writing version."""

    _INSTANCE = PriceAreaAssetWrite


class PriceAreaAssetApplyList(PriceAreaAssetWriteList): ...


def _create_price_area_asset_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    timezone: str | list[str] | None = None,
    timezone_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(timezone, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("timezone"), value=timezone))
    if timezone and isinstance(timezone, list):
        filters.append(dm.filters.In(view_id.as_property_ref("timezone"), values=timezone))
    if timezone_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("timezone"), value=timezone_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections.abc import Sequence
from typing import overload
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.afrr_bid.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.afrr_bid.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    PriceArea,
    PriceAreaFields,
    PriceAreaList,
    PriceAreaTextFields,
)
from cognite.powerops.client._generated.afrr_bid.data_classes._price_area import (
    _PRICEAREA_PROPERTIES_BY_FIELD,
    _create_price_area_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeReadAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .price_area_capacity_price_up import PriceAreaCapacityPriceUpAPI
from .price_area_capacity_price_down import PriceAreaCapacityPriceDownAPI
from .price_area_activation_price_up import PriceAreaActivationPriceUpAPI
from .price_area_activation_price_down import PriceAreaActivationPriceDownAPI
from .price_area_relative_activation import PriceAreaRelativeActivationAPI
from .price_area_total_capacity_allocation_up import PriceAreaTotalCapacityAllocationUpAPI
from .price_area_total_capacity_allocation_down import PriceAreaTotalCapacityAllocationDownAPI
from .price_area_own_capacity_allocation_up import PriceAreaOwnCapacityAllocationUpAPI
from .price_area_own_capacity_allocation_down import PriceAreaOwnCapacityAllocationDownAPI
from .price_area_query import PriceAreaQueryAPI


class PriceAreaAPI(NodeReadAPI[PriceArea, PriceAreaList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[PriceArea]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=PriceArea,
            class_list=PriceAreaList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.capacity_price_up = PriceAreaCapacityPriceUpAPI(client, view_id)
        self.capacity_price_down = PriceAreaCapacityPriceDownAPI(client, view_id)
        self.activation_price_up = PriceAreaActivationPriceUpAPI(client, view_id)
        self.activation_price_down = PriceAreaActivationPriceDownAPI(client, view_id)
        self.relative_activation = PriceAreaRelativeActivationAPI(client, view_id)
        self.total_capacity_allocation_up = PriceAreaTotalCapacityAllocationUpAPI(client, view_id)
        self.total_capacity_allocation_down = PriceAreaTotalCapacityAllocationDownAPI(client, view_id)
        self.own_capacity_allocation_up = PriceAreaOwnCapacityAllocationUpAPI(client, view_id)
        self.own_capacity_allocation_down = PriceAreaOwnCapacityAllocationDownAPI(client, view_id)

    def __call__(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> PriceAreaQueryAPI[PriceAreaList]:
        """Query starting at price areas.

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of price areas to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for price areas.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_price_area_filter(
            self._view_id,
            name,
            name_prefix,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(PriceAreaList)
        return PriceAreaQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more price area.

        Args:
            external_id: External id of the price area to delete.
            space: The space where all the price area are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete price_area by id:

                >>> from cognite.powerops.client._generated.afrr_bid import AFRRBidAPI
                >>> client = AFRRBidAPI()
                >>> client.price_area.delete("my_price_area")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.price_area.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> PriceArea | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> PriceAreaList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> PriceArea | PriceAreaList | None:
        """Retrieve one or more price areas by id(s).

        Args:
            external_id: External id or list of external ids of the price areas.
            space: The space where all the price areas are located.

        Returns:
            The requested price areas.

        Examples:

            Retrieve price_area by id:

                >>> from cognite.powerops.client._generated.afrr_bid import AFRRBidAPI
                >>> client = AFRRBidAPI()
                >>> price_area = client.price_area.retrieve("my_price_area")

        """
        return self._retrieve(external_id, space)

    def search(
        self,
        query: str,
        properties: PriceAreaTextFields | Sequence[PriceAreaTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> PriceAreaList:
        """Search price areas

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of price areas to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Search results price areas matching the query.

        Examples:

           Search for 'my_price_area' in all text properties:

                >>> from cognite.powerops.client._generated.afrr_bid import AFRRBidAPI
                >>> client = AFRRBidAPI()
                >>> price_areas = client.price_area.search('my_price_area')

        """
        filter_ = _create_price_area_filter(
            self._view_id,
            name,
            name_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(self._view_id, query, _PRICEAREA_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PriceAreaFields | Sequence[PriceAreaFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: PriceAreaTextFields | Sequence[PriceAreaTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PriceAreaFields | Sequence[PriceAreaFields] | None = None,
        group_by: PriceAreaFields | Sequence[PriceAreaFields] = None,
        query: str | None = None,
        search_properties: PriceAreaTextFields | Sequence[PriceAreaTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PriceAreaFields | Sequence[PriceAreaFields] | None = None,
        group_by: PriceAreaFields | Sequence[PriceAreaFields] | None = None,
        query: str | None = None,
        search_property: PriceAreaTextFields | Sequence[PriceAreaTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across price areas

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of price areas to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count price areas in space `my_space`:

                >>> from cognite.powerops.client._generated.afrr_bid import AFRRBidAPI
                >>> client = AFRRBidAPI()
                >>> result = client.price_area.aggregate("count", space="my_space")

        """

        filter_ = _create_price_area_filter(
            self._view_id,
            name,
            name_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _PRICEAREA_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: PriceAreaFields,
        interval: float,
        query: str | None = None,
        search_property: PriceAreaTextFields | Sequence[PriceAreaTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for price areas

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of price areas to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_price_area_filter(
            self._view_id,
            name,
            name_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _PRICEAREA_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> PriceAreaList:
        """List/filter price areas

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of price areas to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            List of requested price areas

        Examples:

            List price areas and limit to 5:

                >>> from cognite.powerops.client._generated.afrr_bid import AFRRBidAPI
                >>> client = AFRRBidAPI()
                >>> price_areas = client.price_area.list(limit=5)

        """
        filter_ = _create_price_area_filter(
            self._view_id,
            name,
            name_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._list(limit=limit, filter=filter_)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._commands_config import CommandsConfigApply
    from ._file_ref import FileRefApply
    from ._mapping import MappingApply
    from ._model_template import ModelTemplateApply

__all__ = ["Scenario", "ScenarioApply", "ScenarioList", "ScenarioApplyList", "ScenarioFields", "ScenarioTextFields"]


ScenarioTextFields = Literal["name", "source"]
ScenarioFields = Literal["name", "source"]

_SCENARIO_PROPERTIES_BY_FIELD = {"name": "name", "source": "source"}


class Scenario(DomainModel, protected_namespaces=()):
    space: str = "cogShop"
    name: Optional[str] = None
    model_template: Optional[str] = Field(None, alias="modelTemplate")
    commands: Optional[str] = None
    source: Optional[str] = None
    extra_files: Optional[list[str]] = Field(None, alias="extraFiles")
    mappings_override: Optional[list[str]] = Field(None, alias="mappingsOverride")

    def as_apply(self) -> ScenarioApply:
        return ScenarioApply(
            external_id=self.external_id,
            name=self.name,
            model_template=self.model_template,
            commands=self.commands,
            source=self.source,
            extra_files=self.extra_files,
            mappings_override=self.mappings_override,
        )


class ScenarioApply(DomainModelApply, protected_namespaces=()):
    space: str = "cogShop"
    name: str
    model_template: Union[ModelTemplateApply, str, None] = Field(None, repr=False, alias="modelTemplate")
    commands: Union[CommandsConfigApply, str, None] = Field(None, repr=False)
    source: Optional[str] = None
    extra_files: Union[list[FileRefApply], list[str], None] = Field(default=None, repr=False, alias="extraFiles")
    mappings_override: Union[list[MappingApply], list[str], None] = Field(
        default=None, repr=False, alias="mappingsOverride"
    )

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.name is not None:
            properties["name"] = self.name
        if self.model_template is not None:
            properties["modelTemplate"] = {
                "space": "cogShop",
                "externalId": (
                    self.model_template if isinstance(self.model_template, str) else self.model_template.external_id
                ),
            }
        if self.commands is not None:
            properties["commands"] = {
                "space": "cogShop",
                "externalId": self.commands if isinstance(self.commands, str) else self.commands.external_id,
            }
        if self.source is not None:
            properties["source"] = self.source
        if properties:
            source = dm.NodeOrEdgeData(source=dm.ContainerId("cogShop", "Scenario"), properties=properties)
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space, external_id=self.external_id, existing_version=self.existing_version, sources=sources
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for extra_file in self.extra_files or []:
            edge = self._create_extra_file_edge(extra_file)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(extra_file, DomainModelApply):
                instances = extra_file._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        for mappings_override in self.mappings_override or []:
            edge = self._create_mappings_override_edge(mappings_override)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(mappings_override, DomainModelApply):
                instances = mappings_override._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.model_template, DomainModelApply):
            instances = self.model_template._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.commands, DomainModelApply):
            instances = self.commands._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_extra_file_edge(self, extra_file: Union[str, FileRefApply]) -> dm.EdgeApply:
        if isinstance(extra_file, str):
            end_node_ext_id = extra_file
        elif isinstance(extra_file, DomainModelApply):
            end_node_ext_id = extra_file.external_id
        else:
            raise TypeError(f"Expected str or FileRefApply, got {type(extra_file)}")

        return dm.EdgeApply(
            space="cogShop",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("cogShop", "Scenario.extraFiles"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("cogShop", end_node_ext_id),
        )

    def _create_mappings_override_edge(self, mappings_override: Union[str, MappingApply]) -> dm.EdgeApply:
        if isinstance(mappings_override, str):
            end_node_ext_id = mappings_override
        elif isinstance(mappings_override, DomainModelApply):
            end_node_ext_id = mappings_override.external_id
        else:
            raise TypeError(f"Expected str or MappingApply, got {type(mappings_override)}")

        return dm.EdgeApply(
            space="cogShop",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("cogShop", "Scenario.mappingsOverride"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("cogShop", end_node_ext_id),
        )


class ScenarioList(TypeList[Scenario]):
    _NODE = Scenario

    def as_apply(self) -> ScenarioApplyList:
        return ScenarioApplyList([node.as_apply() for node in self.data])


class ScenarioApplyList(TypeApplyList[ScenarioApply]):
    _NODE = ScenarioApply



from __future__ import annotations

import json
from collections.abc import Iterator
from enum import Enum, auto
from typing import Annotated, Any, ClassVar, Optional

import pandas as pd
from pydantic import BaseModel, constr, validator
from typing_extensions import TypeAlias

from cognite.powerops.prerun_transformations.transformations import Transformation as TransformationV2

ExternalId: TypeAlias = Annotated[str, constr(min_length=1, max_length=255)]


class Auction(str, Enum):
    week = "week"
    weekend = "weekend"


class RetrievalType(Enum):
    RANGE = auto()
    START = auto()
    END = auto()


class AggregationMethod(Enum):
    sum = auto()
    mean = auto()
    std = auto()
    sem = auto()
    max = auto()
    min = auto()
    median = auto()
    first = auto()
    last = auto()


class TransformationType(Enum):
    GATE_SCHEDULE_FLAG_VALUE_MAPPING = auto()
    GENERATOR_PRODUCTION_SCHEDULE_FLAG_VALUE_MAPPING = auto()
    PLANT_PRODUCTION_SCHEDULE_FLAG_VALUE_MAPPING = auto()
    GATE_OPENING_METER_TO_PERCENT = auto()
    TO_BOOL = auto()
    TO_INT = auto()
    ZERO_IF_NOT_ONE = auto()
    MULTIPLY = auto()
    MULTIPLY_FROM_OFFSET = auto()
    ADD = auto()
    ADD_FROM_OFFSET = auto()
    DYNAMIC_ADD_FROM_OFFSET = auto()
    RESERVOIR_LEVEL_TO_VOLUME = auto()
    STATIC = auto()
    DYNAMIC_STATIC = auto()
    ONE_IF_TWO = auto()
    ADD_WATER_IN_TRANSIT = auto()


ATTRIBUTE_DEFAULT_AGGREGATION: dict[str, AggregationMethod | None] = {
    "gate.max_flow": AggregationMethod.min,
    "gate.min_flow": AggregationMethod.max,
    "gate.schedule_flag": AggregationMethod.first,
    "gate.schedule_m3s": AggregationMethod.first,
    "generator.committed_flag": AggregationMethod.first,
    "generator.committed_in": None,
    "generator.initial_state": None,
    "generator.maintenance_flag": AggregationMethod.max,
    "generator.max_p_constr": AggregationMethod.min,
    "generator.min_p_constr": AggregationMethod.max,
    "generator.production_schedule_flag": AggregationMethod.first,
    "market.buy_price": AggregationMethod.mean,
    "market.sale_price": AggregationMethod.mean,
    "plant.feeding_fee": AggregationMethod.mean,
    "plant.maintenance_flag": AggregationMethod.max,
    "plant.max_p_constr": AggregationMethod.min,
    "plant.max_q_constr": AggregationMethod.min,
    "plant.min_p_constr": AggregationMethod.max,
    "plant.mip_flag": AggregationMethod.first,
    "plant.production_schedule": AggregationMethod.first,
    "plant.production_schedule_flag": AggregationMethod.first,
    "reserve_group.rr_up_obligation": AggregationMethod.max,
    "reserve_group.rr_down_obligation": AggregationMethod.max,
    "reservoir.inflow": AggregationMethod.mean,
    "reservoir.max_vol_constr": AggregationMethod.min,
    "reservoir.min_vol_constr": AggregationMethod.max,
    "reservoir.start_head": None,
    "reservoir.tactical_cost_max": AggregationMethod.mean,
    "reservoir.tactical_cost_min": AggregationMethod.mean,
    "reservoir.tactical_limit_max": AggregationMethod.mean,
    "reservoir.tactical_limit_min": AggregationMethod.mean,
}


class Transformation(BaseModel):  # type: ignore[no-redef]  # mypy says kwargs is redefined on this line ¯\_(ツ)_/¯
    transformation: TransformationType
    kwargs: Optional[dict] = None

    @validator("transformation", pre=True)
    def to_type(cls, value):
        return TransformationType[value.upper()] if isinstance(value, str) else value

    def to_dict(self) -> dict:
        return {"transformation": self.transformation.name, "kwargs": self.kwargs}


class TimeSeriesMappingEntryV2(BaseModel):
    object_type: str
    object_name: str
    attribute_name: str
    time_series_external_id: Optional[str] = None
    transformations: Optional[list[TransformationV2]] = None
    retrieve: Optional[RetrievalType] = None
    aggregation: Optional[AggregationMethod] = None

    @validator("aggregation", pre=True)
    def to_enum(cls, value):
        return AggregationMethod[value] if isinstance(value, str) else value

    @validator("aggregation", always=True)
    def set_default(cls, value, values):
        if value is not None:
            return value
        # TODO: do we want default `None` here or raise error?
        return ATTRIBUTE_DEFAULT_AGGREGATION.get(f"{values.get('object_type')}.{values.get('attribute_name')}")

    @validator("retrieve", pre=True)
    def to_retrival_enum(cls, value):
        return RetrievalType[value] if isinstance(value, str) else value

    @property
    def shop_model_path(self) -> str:
        return f"{self.object_type}.{self.object_name}.{self.attribute_name}"

    def _transformations_to_strings(self, max_cols: int) -> list[str]:
        if self.transformations:
            # ensure_ascii=False to allow æøåÆØÅ
            transformation_string = json.dumps([t.to_dict() for t in self.transformations], ensure_ascii=False)
        else:
            transformation_string = json.dumps([])

        max_chars = 255
        return [transformation_string[i * max_chars : (i + 1) * max_chars] for i in range(max_cols)]

    def to_sequence_row(self, max_transformation_cols: int) -> list[str | float]:
        return [
            self.shop_model_path,
            self.time_series_external_id or float("nan"),
            *self._transformations_to_strings(max_cols=max_transformation_cols),
            self.retrieve.name if self.retrieve else float("nan"),
            self.aggregation.name if self.aggregation else float("nan"),
        ]


class TimeSeriesMappingEntry(BaseModel):
    object_type: str
    object_name: str
    attribute_name: str
    time_series_external_id: Optional[str] = None
    transformations: Optional[list[Transformation]] = None
    retrieve: Optional[RetrievalType] = None
    aggregation: Optional[AggregationMethod] = None

    @validator("aggregation", pre=True)
    def to_enum(cls, value):
        return AggregationMethod[value] if isinstance(value, str) else value

    @validator("aggregation", always=True)
    def set_default(cls, value, values):
        if value is not None:
            return value
        # TODO: do we want default `None` here or raise error?
        return ATTRIBUTE_DEFAULT_AGGREGATION.get(f"{values.get('object_type')}.{values.get('attribute_name')}")

    @validator("retrieve", pre=True)
    def to_retrival_enum(cls, value):
        return RetrievalType[value] if isinstance(value, str) else value

    @property
    def shop_model_path(self) -> str:
        return f"{self.object_type}.{self.object_name}.{self.attribute_name}"

    def _transformations_to_strings(self, max_cols: int) -> list[str]:
        if self.transformations:
            # ensure_ascii=False to allow æøåÆØÅ
            transformation_string = json.dumps([t.to_dict() for t in self.transformations], ensure_ascii=False)
        else:
            transformation_string = json.dumps([])

        max_chars = 255
        return [transformation_string[i * max_chars : (i + 1) * max_chars] for i in range(max_cols)]

    def to_sequence_row(self, max_transformation_cols: int) -> list[str | float]:
        return [
            self.shop_model_path,
            self.time_series_external_id or float("nan"),
            *self._transformations_to_strings(max_cols=max_transformation_cols),
            self.retrieve.name if self.retrieve else float("nan"),
            self.aggregation.name if self.aggregation else float("nan"),
        ]


class TimeSeriesMappingV2(BaseModel):
    rows: list[TimeSeriesMappingEntryV2] = []
    columns: ClassVar[list[str]] = [
        "shop_model_path",
        "time_series_external_id",
        "transformations",
        "transformations1",
        "transformations2",
        "transformations3",
        "retrieve",
        "aggregation",
    ]

    @property
    def transformations_cols(self) -> list[str]:
        return [col for col in self.columns if col.startswith("transformations")]

    def __iter__(self) -> Iterator[TimeSeriesMappingEntry]:
        yield from self.rows

    def __len__(self) -> int:
        return len(self.rows)

    def __add__(self, other: TimeSeriesMapping) -> TimeSeriesMapping:
        return TimeSeriesMapping(rows=self.rows + other.rows)

    def append(self, element: TimeSeriesMappingEntry) -> None:
        self.rows.append(element)

    def extend(self, other: TimeSeriesMapping) -> None:
        self.rows.extend(other.rows)

    @property
    def column_definitions(self) -> list[dict]:
        return [{"valueType": "STRING", "externalId": col} for col in self.columns]

    def to_dataframe(self) -> pd.DataFrame:
        rows = [row.to_sequence_row(max_transformation_cols=len(self.transformations_cols)) for row in self.rows]
        return pd.DataFrame(data=rows, columns=self.columns)

    def dumps(self) -> dict[str, Any]:
        rows = []
        for row in self.rows:
            row_raw = row.dict(exclude={"aggregation", "retrieve", "transformations"})
            if row.aggregation:
                row_raw["aggregation"] = row.aggregation.name
            if row.retrieve:
                row_raw["retrieve"] = row.retrieve.name
            if row.transformations:
                row_raw["transformations"] = [
                    {
                        **transformation.dict(exclude={"transformation"}),
                        "transformation": transformation.transformation.name,
                    }
                    for transformation in row.transformations
                ]
            rows.append(row_raw)
        return {"rows": rows}

    @classmethod
    def load_from_dict(cls, config: dict) -> TimeSeriesMappingV2:
        time_series_mappings_v2 = {"rows": []}
        for mapping_entry in config["rows"]:
            if transformations := mapping_entry.get("transformations"):
                loaded_transformations = [TransformationV2.load(t) for t in transformations]
                del mapping_entry["transformations"]
                time_series_mappings_v2["rows"].append(
                    TimeSeriesMappingEntryV2(transformations=loaded_transformations, **mapping_entry)
                )
                continue
            time_series_mappings_v2["rows"].append(TimeSeriesMappingEntryV2(**mapping_entry))

        return cls(**time_series_mappings_v2)


class TimeSeriesMapping(BaseModel):
    rows: list[TimeSeriesMappingEntry] = []
    columns: ClassVar[list[str]] = [
        "shop_model_path",
        "time_series_external_id",
        "transformations",
        "transformations1",
        "transformations2",
        "transformations3",
        "retrieve",
        "aggregation",
    ]

    @property
    def transformations_cols(self) -> list[str]:
        return [col for col in self.columns if col.startswith("transformations")]

    def __iter__(self) -> Iterator[TimeSeriesMappingEntry]:
        yield from self.rows

    def __len__(self) -> int:
        return len(self.rows)

    def __add__(self, other: TimeSeriesMapping) -> TimeSeriesMapping:
        return TimeSeriesMapping(rows=self.rows + other.rows)

    def append(self, element: TimeSeriesMappingEntry) -> None:
        self.rows.append(element)

    def extend(self, other: TimeSeriesMapping) -> None:
        self.rows.extend(other.rows)

    @property
    def column_definitions(self) -> list[dict]:
        return [{"valueType": "STRING", "externalId": col} for col in self.columns]

    def to_dataframe(self) -> pd.DataFrame:
        rows = [row.to_sequence_row(max_transformation_cols=len(self.transformations_cols)) for row in self.rows]
        return pd.DataFrame(data=rows, columns=self.columns)

    def dumps(self) -> dict[str, Any]:
        rows = []
        for row in self.rows:
            row_raw = row.dict(exclude={"aggregation", "retrieve", "transformations"})
            if row.aggregation:
                row_raw["aggregation"] = row.aggregation.name
            if row.retrieve:
                row_raw["retrieve"] = row.retrieve.name
            if row.transformations:
                row_raw["transformations"] = [
                    {
                        **transformation.dict(exclude={"transformation"}),
                        "transformation": transformation.transformation.name,
                    }
                    for transformation in row.transformations
                ]
            rows.append(row_raw)
        return {"rows": rows}



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._commands import Commands, CommandsWrite
    from ._mapping import Mapping, MappingWrite
    from ._model_template import ModelTemplate, ModelTemplateWrite


__all__ = [
    "Scenario",
    "ScenarioWrite",
    "ScenarioApply",
    "ScenarioList",
    "ScenarioWriteList",
    "ScenarioApplyList",
    "ScenarioFields",
    "ScenarioTextFields",
]


ScenarioTextFields = Literal["name", "source"]
ScenarioFields = Literal["name", "source"]

_SCENARIO_PROPERTIES_BY_FIELD = {
    "name": "name",
    "source": "source",
}


class Scenario(DomainModel, protected_namespaces=()):
    """This represents the reading version of scenario.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the scenario.
        data_record: The data record of the scenario node.
        name: The name of the scenario to run
        model_template: The model template to use when running the scenario
        commands: The commands to run
        source: The source of the scenario
        mappings_override: An array of base mappings to override in shop model file
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    model_template: Union[ModelTemplate, str, dm.NodeId, None] = Field(None, repr=False, alias="modelTemplate")
    commands: Union[Commands, str, dm.NodeId, None] = Field(None, repr=False)
    source: Optional[str] = None
    mappings_override: Union[list[Mapping], list[str], None] = Field(default=None, repr=False, alias="mappingsOverride")

    def as_write(self) -> ScenarioWrite:
        """Convert this read version of scenario to the writing version."""
        return ScenarioWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            model_template=(
                self.model_template.as_write() if isinstance(self.model_template, DomainModel) else self.model_template
            ),
            commands=self.commands.as_write() if isinstance(self.commands, DomainModel) else self.commands,
            source=self.source,
            mappings_override=[
                mappings_override.as_write() if isinstance(mappings_override, DomainModel) else mappings_override
                for mappings_override in self.mappings_override or []
            ],
        )

    def as_apply(self) -> ScenarioWrite:
        """Convert this read version of scenario to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ScenarioWrite(DomainModelWrite, protected_namespaces=()):
    """This represents the writing version of scenario.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the scenario.
        data_record: The data record of the scenario node.
        name: The name of the scenario to run
        model_template: The model template to use when running the scenario
        commands: The commands to run
        source: The source of the scenario
        mappings_override: An array of base mappings to override in shop model file
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    model_template: Union[ModelTemplateWrite, str, dm.NodeId, None] = Field(None, repr=False, alias="modelTemplate")
    commands: Union[CommandsWrite, str, dm.NodeId, None] = Field(None, repr=False)
    source: Optional[str] = None
    mappings_override: Union[list[MappingWrite], list[str], None] = Field(
        default=None, repr=False, alias="mappingsOverride"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(Scenario, dm.ViewId("sp_powerops_models", "Scenario", "1"))

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.model_template is not None:
            properties["modelTemplate"] = {
                "space": self.space if isinstance(self.model_template, str) else self.model_template.space,
                "externalId": (
                    self.model_template if isinstance(self.model_template, str) else self.model_template.external_id
                ),
            }

        if self.commands is not None:
            properties["commands"] = {
                "space": self.space if isinstance(self.commands, str) else self.commands.space,
                "externalId": self.commands if isinstance(self.commands, str) else self.commands.external_id,
            }

        if self.source is not None or write_none:
            properties["source"] = self.source

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "Mapping")
        for mappings_override in self.mappings_override or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=mappings_override,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
            )
            resources.extend(other_resources)

        if isinstance(self.model_template, DomainModelWrite):
            other_resources = self.model_template._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.commands, DomainModelWrite):
            other_resources = self.commands._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class ScenarioApply(ScenarioWrite):
    def __new__(cls, *args, **kwargs) -> ScenarioApply:
        warnings.warn(
            "ScenarioApply is deprecated and will be removed in v1.0. Use ScenarioWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "Scenario.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class ScenarioList(DomainModelList[Scenario]):
    """List of scenarios in the read version."""

    _INSTANCE = Scenario

    def as_write(self) -> ScenarioWriteList:
        """Convert these read versions of scenario to the writing versions."""
        return ScenarioWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> ScenarioWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ScenarioWriteList(DomainModelWriteList[ScenarioWrite]):
    """List of scenarios in the writing version."""

    _INSTANCE = ScenarioWrite


class ScenarioApplyList(ScenarioWriteList): ...


def _create_scenario_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    model_template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    source: str | list[str] | None = None,
    source_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if model_template and isinstance(model_template, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("modelTemplate"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": model_template},
            )
        )
    if model_template and isinstance(model_template, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("modelTemplate"),
                value={"space": model_template[0], "externalId": model_template[1]},
            )
        )
    if model_template and isinstance(model_template, list) and isinstance(model_template[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("modelTemplate"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in model_template],
            )
        )
    if model_template and isinstance(model_template, list) and isinstance(model_template[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("modelTemplate"),
                values=[{"space": item[0], "externalId": item[1]} for item in model_template],
            )
        )
    if commands and isinstance(commands, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("commands"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": commands}
            )
        )
    if commands and isinstance(commands, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("commands"), value={"space": commands[0], "externalId": commands[1]}
            )
        )
    if commands and isinstance(commands, list) and isinstance(commands[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("commands"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in commands],
            )
        )
    if commands and isinstance(commands, list) and isinstance(commands[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("commands"),
                values=[{"space": item[0], "externalId": item[1]} for item in commands],
            )
        )
    if isinstance(source, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("source"), value=source))
    if source and isinstance(source, list):
        filters.append(dm.filters.In(view_id.as_property_ref("source"), values=source))
    if source_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("source"), value=source_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from typing import Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
    TimeSeries,
)


__all__ = [
    "SHOPTimeSeries",
    "SHOPTimeSeriesWrite",
    "SHOPTimeSeriesApply",
    "SHOPTimeSeriesList",
    "SHOPTimeSeriesWriteList",
    "SHOPTimeSeriesApplyList",
    "SHOPTimeSeriesFields",
    "SHOPTimeSeriesTextFields",
]


SHOPTimeSeriesTextFields = Literal["object_type", "object_name", "attribute_name", "timeseries"]
SHOPTimeSeriesFields = Literal["object_type", "object_name", "attribute_name", "timeseries"]

_SHOPTIMESERIES_PROPERTIES_BY_FIELD = {
    "object_type": "objectType",
    "object_name": "objectName",
    "attribute_name": "attributeName",
    "timeseries": "timeseries",
}


class SHOPTimeSeries(DomainModel):
    """This represents the reading version of shop time series.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop time series.
        data_record: The data record of the shop time series node.
        object_type: The type of the object
        object_name: The name of the object
        attribute_name: The name of the attribute
        timeseries: Timeseries object from output of SHOP stored as a timeseries in cdf
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "SHOPTimeSeries"
    )
    object_type: Optional[str] = Field(None, alias="objectType")
    object_name: Optional[str] = Field(None, alias="objectName")
    attribute_name: Optional[str] = Field(None, alias="attributeName")
    timeseries: Union[TimeSeries, str, None] = None

    def as_write(self) -> SHOPTimeSeriesWrite:
        """Convert this read version of shop time series to the writing version."""
        return SHOPTimeSeriesWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            object_type=self.object_type,
            object_name=self.object_name,
            attribute_name=self.attribute_name,
            timeseries=self.timeseries,
        )

    def as_apply(self) -> SHOPTimeSeriesWrite:
        """Convert this read version of shop time series to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class SHOPTimeSeriesWrite(DomainModelWrite):
    """This represents the writing version of shop time series.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop time series.
        data_record: The data record of the shop time series node.
        object_type: The type of the object
        object_name: The name of the object
        attribute_name: The name of the attribute
        timeseries: Timeseries object from output of SHOP stored as a timeseries in cdf
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "SHOPTimeSeries"
    )
    object_type: Optional[str] = Field(None, alias="objectType")
    object_name: Optional[str] = Field(None, alias="objectName")
    attribute_name: Optional[str] = Field(None, alias="attributeName")
    timeseries: Union[TimeSeries, str, None] = None

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            SHOPTimeSeries, dm.ViewId("sp_powerops_models", "SHOPTimeSeries", "1")
        )

        properties: dict[str, Any] = {}

        if self.object_type is not None or write_none:
            properties["objectType"] = self.object_type

        if self.object_name is not None or write_none:
            properties["objectName"] = self.object_name

        if self.attribute_name is not None or write_none:
            properties["attributeName"] = self.attribute_name

        if self.timeseries is not None or write_none:
            if isinstance(self.timeseries, str) or self.timeseries is None:
                properties["timeseries"] = self.timeseries
            else:
                properties["timeseries"] = self.timeseries.external_id

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        if isinstance(self.timeseries, CogniteTimeSeries):
            resources.time_series.append(self.timeseries)

        return resources


class SHOPTimeSeriesApply(SHOPTimeSeriesWrite):
    def __new__(cls, *args, **kwargs) -> SHOPTimeSeriesApply:
        warnings.warn(
            "SHOPTimeSeriesApply is deprecated and will be removed in v1.0. Use SHOPTimeSeriesWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "SHOPTimeSeries.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class SHOPTimeSeriesList(DomainModelList[SHOPTimeSeries]):
    """List of shop time series in the read version."""

    _INSTANCE = SHOPTimeSeries

    def as_write(self) -> SHOPTimeSeriesWriteList:
        """Convert these read versions of shop time series to the writing versions."""
        return SHOPTimeSeriesWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> SHOPTimeSeriesWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class SHOPTimeSeriesWriteList(DomainModelWriteList[SHOPTimeSeriesWrite]):
    """List of shop time series in the writing version."""

    _INSTANCE = SHOPTimeSeriesWrite


class SHOPTimeSeriesApplyList(SHOPTimeSeriesWriteList): ...


def _create_shop_time_series_filter(
    view_id: dm.ViewId,
    object_type: str | list[str] | None = None,
    object_type_prefix: str | None = None,
    object_name: str | list[str] | None = None,
    object_name_prefix: str | None = None,
    attribute_name: str | list[str] | None = None,
    attribute_name_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(object_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("objectType"), value=object_type))
    if object_type and isinstance(object_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("objectType"), values=object_type))
    if object_type_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("objectType"), value=object_type_prefix))
    if isinstance(object_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("objectName"), value=object_name))
    if object_name and isinstance(object_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("objectName"), values=object_name))
    if object_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("objectName"), value=object_name_prefix))
    if isinstance(attribute_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("attributeName"), value=attribute_name))
    if attribute_name and isinstance(attribute_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("attributeName"), values=attribute_name))
    if attribute_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("attributeName"), value=attribute_name_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)
from ._bid_method_day_ahead import BidMethodDayAhead, BidMethodDayAheadWrite

if TYPE_CHECKING:
    from ._scenario import Scenario, ScenarioWrite


__all__ = [
    "BidMethodSHOPMultiScenario",
    "BidMethodSHOPMultiScenarioWrite",
    "BidMethodSHOPMultiScenarioApply",
    "BidMethodSHOPMultiScenarioList",
    "BidMethodSHOPMultiScenarioWriteList",
    "BidMethodSHOPMultiScenarioApplyList",
    "BidMethodSHOPMultiScenarioFields",
    "BidMethodSHOPMultiScenarioTextFields",
]


BidMethodSHOPMultiScenarioTextFields = Literal[
    "name", "shop_start_specification", "shop_end_specification", "shop_bid_date_specification"
]
BidMethodSHOPMultiScenarioFields = Literal[
    "name", "shop_start_specification", "shop_end_specification", "shop_bid_date_specification"
]

_BIDMETHODSHOPMULTISCENARIO_PROPERTIES_BY_FIELD = {
    "name": "name",
    "shop_start_specification": "shopStartSpecification",
    "shop_end_specification": "shopEndSpecification",
    "shop_bid_date_specification": "shopBidDateSpecification",
}


class BidMethodSHOPMultiScenario(BidMethodDayAhead):
    """This represents the reading version of bid method shop multi scenario.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid method shop multi scenario.
        data_record: The data record of the bid method shop multi scenario node.
        name: Name for the BidMethod
        shop_start_specification: The shop start specification
        shop_end_specification: The shop end specification
        shop_bid_date_specification: The shop bid date specification
        scenarios: The scenarios to run this bid method with (includes incremental mappings and base mappings)
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "BidMethodSHOPMultiScenario"
    )
    shop_start_specification: Optional[str] = Field(None, alias="shopStartSpecification")
    shop_end_specification: Optional[str] = Field(None, alias="shopEndSpecification")
    shop_bid_date_specification: Optional[str] = Field(None, alias="shopBidDateSpecification")
    scenarios: Union[list[Scenario], list[str], None] = Field(default=None, repr=False)

    def as_write(self) -> BidMethodSHOPMultiScenarioWrite:
        """Convert this read version of bid method shop multi scenario to the writing version."""
        return BidMethodSHOPMultiScenarioWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            shop_start_specification=self.shop_start_specification,
            shop_end_specification=self.shop_end_specification,
            shop_bid_date_specification=self.shop_bid_date_specification,
            scenarios=[
                scenario.as_write() if isinstance(scenario, DomainModel) else scenario
                for scenario in self.scenarios or []
            ],
        )

    def as_apply(self) -> BidMethodSHOPMultiScenarioWrite:
        """Convert this read version of bid method shop multi scenario to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidMethodSHOPMultiScenarioWrite(BidMethodDayAheadWrite):
    """This represents the writing version of bid method shop multi scenario.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid method shop multi scenario.
        data_record: The data record of the bid method shop multi scenario node.
        name: Name for the BidMethod
        shop_start_specification: The shop start specification
        shop_end_specification: The shop end specification
        shop_bid_date_specification: The shop bid date specification
        scenarios: The scenarios to run this bid method with (includes incremental mappings and base mappings)
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "BidMethodSHOPMultiScenario"
    )
    shop_start_specification: Optional[str] = Field(None, alias="shopStartSpecification")
    shop_end_specification: Optional[str] = Field(None, alias="shopEndSpecification")
    shop_bid_date_specification: Optional[str] = Field(None, alias="shopBidDateSpecification")
    scenarios: Union[list[ScenarioWrite], list[str], None] = Field(default=None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            BidMethodSHOPMultiScenario, dm.ViewId("sp_powerops_models", "BidMethodSHOPMultiScenario", "1")
        )

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.shop_start_specification is not None or write_none:
            properties["shopStartSpecification"] = self.shop_start_specification

        if self.shop_end_specification is not None or write_none:
            properties["shopEndSpecification"] = self.shop_end_specification

        if self.shop_bid_date_specification is not None or write_none:
            properties["shopBidDateSpecification"] = self.shop_bid_date_specification

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "BidMethodDayahead.scenarios")
        for scenario in self.scenarios or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache, start_node=self, end_node=scenario, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        return resources


class BidMethodSHOPMultiScenarioApply(BidMethodSHOPMultiScenarioWrite):
    def __new__(cls, *args, **kwargs) -> BidMethodSHOPMultiScenarioApply:
        warnings.warn(
            "BidMethodSHOPMultiScenarioApply is deprecated and will be removed in v1.0. Use BidMethodSHOPMultiScenarioWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BidMethodSHOPMultiScenario.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BidMethodSHOPMultiScenarioList(DomainModelList[BidMethodSHOPMultiScenario]):
    """List of bid method shop multi scenarios in the read version."""

    _INSTANCE = BidMethodSHOPMultiScenario

    def as_write(self) -> BidMethodSHOPMultiScenarioWriteList:
        """Convert these read versions of bid method shop multi scenario to the writing versions."""
        return BidMethodSHOPMultiScenarioWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BidMethodSHOPMultiScenarioWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidMethodSHOPMultiScenarioWriteList(DomainModelWriteList[BidMethodSHOPMultiScenarioWrite]):
    """List of bid method shop multi scenarios in the writing version."""

    _INSTANCE = BidMethodSHOPMultiScenarioWrite


class BidMethodSHOPMultiScenarioApplyList(BidMethodSHOPMultiScenarioWriteList): ...


def _create_bid_method_shop_multi_scenario_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    shop_start_specification: str | list[str] | None = None,
    shop_start_specification_prefix: str | None = None,
    shop_end_specification: str | list[str] | None = None,
    shop_end_specification_prefix: str | None = None,
    shop_bid_date_specification: str | list[str] | None = None,
    shop_bid_date_specification_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(shop_start_specification, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("shopStartSpecification"), value=shop_start_specification)
        )
    if shop_start_specification and isinstance(shop_start_specification, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("shopStartSpecification"), values=shop_start_specification)
        )
    if shop_start_specification_prefix is not None:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("shopStartSpecification"), value=shop_start_specification_prefix)
        )
    if isinstance(shop_end_specification, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopEndSpecification"), value=shop_end_specification))
    if shop_end_specification and isinstance(shop_end_specification, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopEndSpecification"), values=shop_end_specification))
    if shop_end_specification_prefix is not None:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("shopEndSpecification"), value=shop_end_specification_prefix)
        )
    if isinstance(shop_bid_date_specification, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("shopBidDateSpecification"), value=shop_bid_date_specification)
        )
    if shop_bid_date_specification and isinstance(shop_bid_date_specification, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("shopBidDateSpecification"), values=shop_bid_date_specification)
        )
    if shop_bid_date_specification_prefix is not None:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("shopBidDateSpecification"), value=shop_bid_date_specification_prefix
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from cognite.client import data_modeling as dm, CogniteClient

from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    PriceAreaAsset,
)
from ._core import DEFAULT_QUERY_LIMIT, QueryBuilder, QueryStep, QueryAPI, T_DomainModelList, _create_edge_filter

if TYPE_CHECKING:
    from .plant_query import PlantQueryAPI
    from .watercourse_query import WatercourseQueryAPI


class PriceAreaAssetQueryAPI(QueryAPI[T_DomainModelList]):
    def __init__(
        self,
        client: CogniteClient,
        builder: QueryBuilder[T_DomainModelList],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId],
        filter_: dm.filters.Filter | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
    ):
        super().__init__(client, builder, view_by_read_class)

        self._builder.append(
            QueryStep(
                name=self._builder.next_name("price_area_asset"),
                expression=dm.query.NodeResultSetExpression(
                    from_=self._builder[-1].name if self._builder else None,
                    filter=filter_,
                ),
                select=dm.query.Select([dm.query.SourceSelector(self._view_by_read_class[PriceAreaAsset], ["*"])]),
                result_cls=PriceAreaAsset,
                max_retrieve_limit=limit,
            )
        )

    def plants(
        self,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
    ) -> PlantQueryAPI[T_DomainModelList]:
        """Query along the plant edges of the price area asset.

        Args:
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plant edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            PlantQueryAPI: The query API for the plant.
        """
        from .plant_query import PlantQueryAPI

        from_ = self._builder[-1].name

        edge_filter = _create_edge_filter(
            dm.DirectRelationReference("sp_powerops_types", "isPlantOf"),
            external_id_prefix=external_id_prefix,
            space=space,
        )
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("plants"),
                expression=dm.query.EdgeResultSetExpression(
                    filter=edge_filter,
                    from_=from_,
                    direction="outwards",
                ),
                select=dm.query.Select(),
                max_retrieve_limit=limit,
            )
        )
        return PlantQueryAPI(self._client, self._builder, self._view_by_read_class, None, limit)

    def watercourses(
        self,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
    ) -> WatercourseQueryAPI[T_DomainModelList]:
        """Query along the watercourse edges of the price area asset.

        Args:
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of watercourse edges to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            WatercourseQueryAPI: The query API for the watercourse.
        """
        from .watercourse_query import WatercourseQueryAPI

        from_ = self._builder[-1].name

        edge_filter = _create_edge_filter(
            dm.DirectRelationReference("sp_powerops_types", "isWatercourseOf"),
            external_id_prefix=external_id_prefix,
            space=space,
        )
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("watercourses"),
                expression=dm.query.EdgeResultSetExpression(
                    filter=edge_filter,
                    from_=from_,
                    direction="outwards",
                ),
                select=dm.query.Select(),
                max_retrieve_limit=limit,
            )
        )
        return WatercourseQueryAPI(self._client, self._builder, self._view_by_read_class, None, limit)

    def query(
        self,
    ) -> T_DomainModelList:
        """Execute query and return the result.

        Returns:
            The list of the source nodes of the query.

        """
        return self._query()
